#include "jpegreader.h"
#include "md5.h"
#include <stdlib.h>

JPEGReader::JPEGReader()
{
    m_version = 0;  
    m_date = 1;
    m_data = NULL;
    m_reader = NULL;
}

JPEGReader::~JPEGReader()
{
    if(m_data != NULL)
        delete []m_data;    
    if(m_reader != NULL)
        delete m_reader;    
}

void JPEGReader::scanJPEGHeader(const char * filename)
{
    setStatus(false);

    FileReader *reader = new FileReader(filename);
            
    if(!reader->checkStatus())
        error("Error: can not open %s input file.\n", filename);

    setFileReader(reader);

    int marker;
    
    /* Expect SOI at start of file */
    if(getFirstMarker() != M_SOI)
        error("Error: can not get SOI marker.\n");
    
    /* Scan miscellaneous markers until we reach SOS. */
    bool finished = false;
    while(!finished) 
    {
        marker = getNextMarker();
        switch (marker) 
        {
        case M_SOS:         /* stop before hitting compressed data */
            finished = true;
            break;
            
        case M_EOI:         /* in case it's a tables-only JPEG stream */
            finished = true;
            break;
            
        case M_APP7:
            processMarker();
            setStatus(true);
            finished = true;    
            break;
        default:           /* Anything else just gets skipped */
            skipData();    /* we assume it has a parameter count... */
            break;
        }
    } 

    setFileReader(NULL);

    if(m_data != NULL)
    {
        FileReader *rd = new FileReader(filename);
        long pos = 0;
        int ret;
        unsigned char c;
        MD5 md;
        md.init();
        
        while((ret = rd->getNextByte(c)) == 1)
        {
            if((pos < m_crc_pos)||(pos > m_crc_pos + 16))
                md.update(&c, 1);
            pos++;
        }
        md.done();          
        delete rd;
        
        for(int i = 0; i < 16; i++)
            if(md.digest[i] != m_crc[i])
                error("Error: CRC check failed\n");
    }
}

unsigned char JPEGReader::read1Byte()
{
    unsigned char c;
    int res;
    res = m_reader->getNextByte(c);

    if(res != 1)
        error("Error: premature EOF in JPEG file\n");

    return c;
}

unsigned int JPEGReader::read2Bytes()
{
    unsigned char c[2];
    for(int i = 0; i < 2; i++)
        c[i] = read1Byte();         

    return (((unsigned int) c[0]) << 8) + ((unsigned int) c[1]);
}

unsigned long JPEGReader::read4Bytes()
{
    unsigned char c[4];
    for(int i = 0; i < 4; i++)
        c[i] = read1Byte();         

    return (((unsigned long) c[0]) << 24) + (((unsigned long) c[1]) << 16) + 
            (((unsigned long) c[2]) << 8) + ((unsigned long) c[3]);
}

int JPEGReader::getFirstMarker()
{
    unsigned char c1, c2;
    
    c1 = read1Byte();
    c2 = read1Byte();
    if(c1 != 0xFF || c2 != M_SOI)
        error("Error: not a JPEG file");
    return c2;
}

int JPEGReader::getNextMarker()
{
    int c;
    int discarded_bytes = 0;
    
    c = read1Byte();
    while (c != 0xFF) 
    {
        discarded_bytes++;
        c = read1Byte();
    }
    
    do 
    {
        c = read1Byte();
    } while (c == 0xFF);
    
    if(discarded_bytes != 0) 
        warning("Waring: garbage data found in JPEG file\n");
    
    return c;
}

void JPEGReader::skipData()
{
    unsigned int length;
    
    /* Get the marker parameter length count */
    length = read2Bytes();
    /* Length includes itself, so must be at least 2 */
    if (length < 2)
        error("Error: wrong JPEG marker length\n");

    length -= 2;
    /* Skip over the remaining bytes */
    while (length > 0) 
    {
        read1Byte();
        length--;
    }
}

void JPEGReader::processMarker()
{
    unsigned int length = read2Bytes();
    if (length < 2)
        error("Error: wrong JPEG marker length\n");

    length -= 2;

    unsigned int i, header_length = 0;

    unsigned int magic = read2Bytes();
    if(magic != MAGIC)
        error("Error: wrong marker header format\n");   
    
    header_length = read2Bytes();

    m_version = read1Byte();
    if(m_version > CURRENT_VERSION) 
        warning("Waring: supported format version %d is less that file format version %d\n", CURRENT_VERSION, m_version);

    if((m_version == VERSION_1)&&(header_length != VERSION_1_HEADER_SIZE))  
        error("Error: wrong marker header length\n");

    m_date = read4Bytes();

    m_crc_pos = m_reader->getCurrentPosition();

    for(i = 0; i < 16; i++)
        m_crc[i] = read1Byte();

    for(i = 0; i < header_length - VERSION_1_HEADER_SIZE; i++) read1Byte();

    m_length = length - header_length - 2;

    m_data = new unsigned char[m_length];

    for(i = 0 ;i < m_length; i++) 
        m_data[i] = (unsigned char)read1Byte();
}


void JPEGReader::setFileReader(FileReader *reader) 
{ 
    if(m_reader != NULL)
        delete m_reader;
    m_reader = reader;
}

void JPEGReader::printCRC(char * buff)
{
    char * cur = buff;
    for(int i = 0; i < 16; i++)
    {
        printByte(m_crc[i],cur);
        cur += 2;
    }
    *cur = 0;
}

void JPEGReader::printByte(unsigned char byte, char * buff)
{
    unsigned char n1 = byte >> 4;
    unsigned char n2 = byte % 16;

    printSymbol(n1, buff);
    printSymbol(n2, buff + 1);
}

void JPEGReader::printSymbol(unsigned char n, char * buff)
{
    if(n > 9)
        *buff = 'A' + (n - 10);
    else
        *buff = '0' + n;
}
