#include "jc.h"
#include "jpegwriter.h"
#include "filewriter.h"
#include "md5.h"

#define MAX_BUFFER_SIZE 65533
#define BUFFER_SIZE (MAX_BUFFER_SIZE + 1)

JPEGWriter::JPEGWriter()
{
    m_dataBuffer = NULL;
    m_dataSize = 0;
}

JPEGWriter::~JPEGWriter()
{
    if(m_dataBuffer != NULL) 
        delete []m_dataBuffer;
}

void JPEGWriter::writeJPEGMarkerContent(char *src_filename, char *data_filename, char *tgt_filename)
{
    setStatus(false);

    scanInputFile(src_filename);

    if(data_filename != NULL)
        readDataFile(data_filename);

    createNewFile(src_filename, tgt_filename);

    setStatus(true);
}

void JPEGWriter::scanInputFile(char * filename)
{
    m_appMarkerStart[0] = 0;
    m_appMarkerFinish[0] = 0;
    m_appMarkerStart[1] = 0;
    m_appMarkerFinish[1] = 0;

    FileReader *reader = new FileReader(filename);
            
    if(!reader->checkStatus())
        error("Error: can not open %s input file.\n", filename);

    setFileReader(reader);

    int marker;
    
    /* Expect SOI at start of file */
    if(getFirstMarker() != M_SOI)
        error("Error: can not get SOI marker.\n");
    
    /* Scan miscellaneous markers until we reach SOS. */
    for (;;) 
    {
        marker = getNextMarker();
        switch (marker) 
        {
        case M_SOS:         /* stop before hitting compressed data */
            return;
            
        case M_EOI:         /* in case it's a tables-only JPEG stream */
            return;

        case M_APP0:            
            processMarker(marker);
            break;

        case M_APP7:
            processMarker(marker);
            break;
        default:           /* Anything else just gets skipped */
            skipData();    /* we assume it has a parameter count... */
            break;
        }
    } 

    setFileReader(NULL);    
}

void JPEGWriter::processMarker(int marker)
{
    int idx = (marker == M_APP0)?0:1;
    m_appMarkerStart[idx] = getFileReader()->getCurrentPosition();

    unsigned int length = read2Bytes();
    if (length < 2)
        error("Error: wrong JPEG marker length\n");

    length -= 2;

    for(unsigned int i = 0 ;i < length; i++) 
        read1Byte();

    m_appMarkerFinish[idx] = getFileReader()->getCurrentPosition();
    if(m_appMarkerFinish[idx] <= m_appMarkerStart[idx] + 2)
        error("Error: Bad marker %x format. Position %ld",marker, m_appMarkerStart[idx]);
}

void JPEGWriter::readDataFile(char * filename)
{
    int handle = open(filename, O_RDONLY | O_BINARY);

    if(handle < 0)
        error("Error: can not open data file %s",filename);

    m_dataBuffer = new unsigned char[BUFFER_SIZE];
    m_dataSize = read( handle , m_dataBuffer , BUFFER_SIZE );

    if(m_dataSize > MAX_BUFFER_SIZE)
    {
        warning("Warning: can not store more than %ld bytes from %s file.\n",MAX_BUFFER_SIZE,filename); 
        m_dataSize = MAX_BUFFER_SIZE;
    }

    close(handle);
}

void JPEGWriter::createNewFile(char *src_filename, char *tgt_filename)
{
    bool hasData = (m_dataSize != 0);

    int tgt_handle = creat(tgt_filename, S_IREAD | S_IWRITE);
    close(tgt_handle);
    tgt_handle = open(tgt_filename, O_WRONLY | O_BINARY);
    if(tgt_handle < 0)
        error("Error: can not create output file %s",tgt_filename);

    int src_handle = open(src_filename, O_RDONLY | O_BINARY);
    if(src_handle < 0)
        error("Error: can not open %s input file.\n", src_filename);
    
    long crc_pos = -1;
    if(m_appMarkerStart[1] > 0)
    {
        copyData(0,m_appMarkerStart[1] - 2, src_handle, tgt_handle);
        if(hasData) crc_pos = writeMarker(tgt_handle);
        copyData(m_appMarkerFinish[1], -1, src_handle, tgt_handle);
    }
    else
    {
        copyData(0, m_appMarkerFinish[0], src_handle, tgt_handle);
        if(hasData) crc_pos = writeMarker(tgt_handle);
        copyData(m_appMarkerFinish[0], -1, src_handle, tgt_handle);
    }

    close(src_handle);
    close(tgt_handle);

    if(hasData)
    {
        FileReader *reader = new FileReader(tgt_filename);
        long pos = 0;
        int ret;
        unsigned char c;
        MD5 md;
        md.init();
        
        while((ret = reader->getNextByte(c)) == 1)
        {
            if((pos < crc_pos)||(pos > crc_pos + 16))
                md.update(&c, 1);
            pos++;
        }
        md.done();          
        
        delete reader;
        
        tgt_handle = open(tgt_filename, O_WRONLY | O_BINARY);
        lseek(tgt_handle,crc_pos,SEEK_SET);
        write(tgt_handle, md.digest, 16);
        close(tgt_handle);
    }
}

void JPEGWriter::copyData(long start_pos, long finish_pos, int src_handle, int tgt_handle)
{
    FileReader *reader = new FileReader(src_handle);
    FileWriter *writer = new FileWriter(tgt_handle); 
    unsigned char c;
    int ret,ret2;

    lseek(src_handle,start_pos,SEEK_SET);

    if(finish_pos == -1)
    {
        while((ret = reader->getNextByte(c)) == 1)
        {
            ret2 = writer->setNextByte(c);
            if(ret2 != 1)
                error("Error: can not write target file data");
        }

        if(ret == -1)
            error("Error: can not read source file data");
    }
    else
    {
        for(long i = start_pos; i < finish_pos; i++)
        {
            ret = reader->getNextByte(c);
            if(ret != 1)
                error("Error: can not read source file data");
            ret2 = writer->setNextByte(c);
            if(ret2 != 1)
                error("Error: can not write target file data");
        }
    }

    delete writer;
    delete reader;
}

long JPEGWriter::writeMarker(int handle)
{
    write1byte(0xFF, handle);
    write1byte(M_APP7, handle);
    write2bytes(2 + 2 + VERSION_1_HEADER_SIZE + m_dataSize, handle);
    write2bytes(MAGIC, handle);
    write2bytes(VERSION_1_HEADER_SIZE, handle);
    write1byte(CURRENT_VERSION, handle);
    write4bytes((unsigned long)time(NULL), handle);
    int crc_pos = lseek(handle,0,SEEK_CUR);
    unsigned char crc[16] = {7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7};
    write(handle, crc, 16);
    write(handle,m_dataBuffer,m_dataSize);

    return crc_pos;
}

void JPEGWriter::write1byte(unsigned char data, int handle)
{
    write(handle, &data, 1);
}

void JPEGWriter::write2bytes(unsigned int data, int handle)
{
    unsigned char b1 = (unsigned char)(data & 0xFF);
    unsigned char b2 = (unsigned char)(data >> 8);
    write1byte(b2,handle);
    write1byte(b1,handle);
}

void JPEGWriter::write4bytes(unsigned long data, int handle)
{
    unsigned int i1 = (unsigned int)(data & 0xFFFF);
    unsigned int i2 = (unsigned int)(data >> 16);
    write2bytes(i2, handle);
    write2bytes(i1, handle);
}

